import os
import shutil
from filecmp import dircmp

from sonotoria.jinja import template_folder, template_file

TEST_DEST = 'tests/test_tmp'

def reset_dest_folder():
    if os.path.exists(TEST_DEST):
        shutil.rmtree(TEST_DEST)
    os.mkdir(TEST_DEST)

def use_test_dest(ftest):
    reset_dest_folder()
    ftest()
    reset_dest_folder()

def same_res(res_folder):
    comp = dircmp(TEST_DEST, res_folder)
    assert len(comp.left_only) == 0
    assert len(comp.right_only) == 0
    assert len(comp.diff_files) == 0

@use_test_dest
def test_templating_file():
    # Given
    src = 'tests/test_files/src/file'
    dest = f'{TEST_DEST}/test'

    # When
    template_file(src, dest, context={'foo': 'bar'})

    # Then
    with open(dest, encoding='utf-8') as fd:
        assert fd.read() == 'bar'

@use_test_dest
def test_templating_file_no_path_in_dest():
    # Given
    src = 'tests/test_files/src/file'
    dest = '__test'

    # When
    template_file(src, dest, context={'foo': 'bar'})

    # Then
    with open(dest, encoding='utf-8') as fd:
        assert fd.read() == 'bar'
    os.system('rm test')

@use_test_dest
def test_templating_default_values():
    # Given
    src = 'tests/test_files/src'

    # When
    template_folder(src, TEST_DEST)

    # Then
    same_res('tests/test_files/res_default')

@use_test_dest
def test_templating_overridden_names():
    # Given
    src = 'tests/test_files/src'

    # When
    template_folder(src, TEST_DEST, context={'male_names':['chaest', 'polkka'], 'female_names':[]})

    # Then
    same_res('tests/test_files/res_override')

@use_test_dest
def test_templating_no_config():
    # Given
    src = 'tests/test_files/no_config'

    # When
    template_folder(src, TEST_DEST, context={'value': 'test'})

    # Then
    same_res('tests/test_files/res_no_config')

@use_test_dest
def test_templating_two_dockerfiles():
    # Given
    src = 'tests/test_files/two_dockerfiles'

    # When
    template_folder(src, TEST_DEST, context={'value': 'test'})

    # Then
    same_res('tests/test_files/res_two_dockerfiles')



# -*- coding: utf-8 -*-
import os
import logging
from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader, module_from_spec

import yaml

from . import jaml

logger = logging.getLogger(__name__)


class WrongConfigException(Exception):
    pass

template_path = lambda src: f'{src}/.template'
config_path = lambda src: f'{template_path(src)}/config.yaml'
default_path = lambda src: f'{template_path(src)}/default.yaml'
vars_path = lambda src: f'{template_path(src)}/vars.yaml'

def load_env(src, context, filters, tests):
    config  = load_config(src)
    filters = load_funcs(src, 'filters', filters, context)
    tests   = load_funcs(src, 'tests', tests, context)
    context = load_context(src, config, context, filters, tests)

    excluded = config.get('exclude', []) + ['.template']
    return {
        'context': context,
        'filters': filters,
        'tests': tests,
        'excluded': excluded,
        'excluded_paths': [f'{src}/{excluded_path}' for excluded_path in excluded],
        'config': config
    }

def load_config(src):
    if not os.path.exists(template_path(src)):
        logger.warning('No `.template` folder found, using empty config.')
        return {}

    if not os.path.exists(config_path(src)):
        logger.warning('No config file found, using empty config.')
        return {}

    with open(config_path(src), encoding='utf-8') as config_f:
        return yaml.safe_load(config_f)

def load_context(src, config, context, filters, tests):
    config_default = config.get('default', {})

    if not os.path.exists(default_path(src)):
        logger.warning('No default file found.')
        default_values = config_default
    else:
        with open(default_path(src), encoding='utf-8') as defaults_f:
            default_values = yaml.safe_load(defaults_f)
        default_values.update(config_default)

    if default_values == {}:
        logger.warning('No default values provided.')

    new_context = { **default_values, **(context or {}) }

    if not os.path.exists(vars_path(src)):
        logger.warning('No vars file found.')
    else:
        new_context = {
            **new_context,
            **jaml.load(
                vars_path(src),
                context=new_context,
                filters=filters,
                tests=tests
            )
        }

    return new_context


def unified_loop_config(config):
    ensure_folder_or_file(config)

    type_, path = get_type_and_path(config)

    if 'var' not in config:
        raise WrongConfigException('A `var` parameter is required for a loop config.')

    if 'item' not in config:
        if not config['var'].endswith('s'):
            raise WrongConfigException('The `var` parameter should end with an `s` if no `item` name is provided.')
        item = config['var'][:-1]
    else:
        item = config['item']

    transform = get_transform(path, config, item)

    return {
        'type': type_,
        'path': path,
        'transform': transform,
        'var': config['var'],
        'item': item,
        'excluded_values': config.get('exclude_values', []),
        'filters': config.get('filters', []),
        'tests': config.get('when', []),
        'loop': config.get('loop', []),
        'rename': config.get('rename', [])
    }

def unified_rename_config(config):
    ensure_folder_or_file(config)

    type_, path = get_type_and_path(config)

    if 'var' not in config and 'transform' not in config:
        logger.warning(f'Neither `var` nor `transform` were provided. No transformation to {type_} at {path}.')

    if 'var' in config and 'transform' in config:
        logger.warning('A `var` was provided together with a `transform` in a rename config. Only the latter will be used.')

    transform = get_transform(path, config, config.get('var', None))

    common_config = {
        'type': type_,
        'path': path,
        'transform': transform
    }

    return common_config if type_ == 'file' else { **common_config,
        'loop': config.get('loop', []),
        'rename': config.get('rename', [])
    }

def ensure_folder_or_file(config):
    if 'folder' in config and 'file' in config:
        raise WrongConfigException('A config cannot have both a `folder` or `file`.')
    if 'folder' not in config and 'file' not in config:
        raise WrongConfigException('A config cannot have neither a `folder` nor `file`.')

def get_type_and_path(config):
    type_ = 'folder' if 'folder' in config else 'file'
    path = config[type_]

    if type_ == 'file' and ('rename' in config or 'loop' in config):
        raise WrongConfigException(f'A file loop config cannot have a nested rename or loop (@ {path}).')

    return type_, path

def get_transform(path, config, var):
    transform = config.get('transform', path.split('/')[-1])
    if 'var' in config and 'transform' not in config:
        ext = '.'.join(transform.split('.')[1:])
        if len(ext) > 0:
            ext = '.' + ext
        transform = '{{ ' + var + ' }}' + ext
    return transform

def load_funcs(src, type_, existing_funcs, context):
    module = load_funcs_module(src, type_)
    funcs = { **getattr(module, f'get_{type_}', lambda _: {})(context), **(existing_funcs or {}) }
    if len(funcs) == 0:
        logger.warning(f'No {type_} loaded.')
    return funcs

def load_funcs_module(src, type_):
    path_to_funcs = f'{template_path(src)}/{type_}.py'
    module_name = f'{type_}_specs'

    if not os.path.exists(path_to_funcs):
        logger.warning(f'No {type_} file found.')
        return None

    loader = SourceFileLoader(module_name, path_to_funcs)
    spec = spec_from_loader(module_name, loader)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



# -*- coding: utf-8 -*-
import os
import shutil
import logging
from typing import Any, Dict, Callable

import jinja2
from benedict import benedict

from .template_config import load_env, unified_loop_config, unified_rename_config, WrongConfigException

logger = logging.getLogger(__name__)

def template_string(
    str_: str,
    context: Dict[str, Any]                   = None,
    filters: Dict[str, Callable[[Any], Any]]  = None,
    tests:   Dict[str, Callable[[Any], bool]] = None
) -> str:
    """This function templates a given string using jinja2 engine.

    Args:
        str_: The string to template.
        context: A dict with the context of the template.
        filters: A dict with filters to add to the template engine. Names as keys, functions as values.
        tests: A dict of tests to add to the template engine. Names as keys, functions as values.

    Returns:
        The templated string

    Examples:
       >>> from sonotoria import jinja
       >>> jinja.template_string('my {{ name }} is rich', context={'name': 'tailor'})
       my tailor is rich
    """
    context = context or {}

    # Somehow trailing line feed are removed by jinja2 readding them
    add_trail = '\n' if str_.endswith('\n') else ''

    env = jinja2.Environment()
    for name, filter_ in (filters or {}).items():
        env.filters[name] = filter_
    for name, test in (tests or {}).items():
        env.tests[name] = test
    try:
        return env.from_string(str_).render(**context) + add_trail
    except TypeError: # Yaml Objects
        return env.from_string(str_).render(**context.__dict__) + add_trail

def template_file(
    src: str,
    dest: str,
    context: Dict[str, Any]                   = None,
    filters: Dict[str, Callable[[Any], Any]]  = None,
    tests:   Dict[str, Callable[[Any], bool]] = None
) -> str:
    """This function templates a given file using jinja2 engine.

    Args:
        src: The path to the file to template.
        dest: The path where to create the resulting templated file.
        context: A dict with the context of the template.
        filters: A dict with filters to add to the template engine. Names as keys, functions as values.
        tests: A dict of tests to add to the template engine. Names as keys, functions as values.

    Returns:
        Nothing

    Examples:
       >>> from sonotoria import jinja
       >>> jinja.template_file('mytemplate', 'templated', context={'name': 'tailor'})
    """
    with open(src, 'r', encoding='utf-8') as src_f:
        content = template_string(src_f.read(), context, filters, tests)

    if '/' in dest:
        os.makedirs(os.path.dirname(dest), exist_ok=True)

    with open(dest, 'w', encoding='utf-8') as dest_f:
        dest_f.write(content)

def template_folder(
    src: str,
    dest: str,
    context: Dict[str, Any]                   = None,
    filters: Dict[str, Callable[[Any], Any]]  = None,
    tests:   Dict[str, Callable[[Any], bool]] = None
) -> str:
    """This function templates a given folder using jinja2 engine and configuration provided in this folder. Examples found in README.

    Args:
        src: The path to the folder to template.
        dest: The path where to create the resulting templated folder.
        context: A dict with the context of the template.
        filters: A dict with filters to add to the template engine. Names as keys, functions as values.
        tests: A dict of tests to add to the template engine. Names as keys, functions as values.

    Returns:
        Nothing

    Examples:
       >>> from sonotoria import jinja
       >>> jinja.template_folder('mytemplate', 'templated', context={'name': 'tailor'})
    """
    env = load_env(src, context, filters, tests)

    if 'template' not in env['config']:
        logger.warning('No template config found.')
        handle_folder(src, dest, env)
    else:
        handle_node(env['config']['template'], src, dest, env)

def without_excluded(env):
    return {
        'context': env['context'],
        'filters': env['filters'],
        'tests': env['tests']
    }

def handle_node(node_config, src, dest, env):
    handled_paths = []
    for loop in node_config.get('loop', []):
        handled_paths += handle_loop(loop, src, dest, env)
    for rename in node_config.get('rename', []):
        handled_paths += handle_rename(rename, src, dest, env)

    handled_paths += handle_folder(src, dest, env, handled_paths)
    return handled_paths

def handle_folder(src, dest, env, handled_paths=None):
    newly_handled_path = []
    paths_to_handle = unhandled_paths(src, handled_paths or [], env.get('excluded_paths', []))
    for path in paths_to_handle:
        path_dest = path.replace(src, dest)
        if os.path.isdir(path):
            os.mkdir(path_dest)
        if os.path.isfile(path):
            try:
                template_file(path, path_dest, **without_excluded(env))
            except UnicodeDecodeError as err:
                print(f'Could not template {path} -> {path_dest} due to non unicode encoding.')
                print(err)
                print('File will be copied as is.')
                os.makedirs(os.path.dirname(path_dest), exist_ok=True)
                shutil.copy(path, path_dest)
        newly_handled_path.append(path)
    return newly_handled_path

def unhandled_paths(root, handled_paths, excluded_paths):
    paths = []
    for rootdir, dirs, files in os.walk(root):
        for file_ in files:
            paths.append(os.path.join(rootdir, file_))
        for subdir in dirs:
            paths.append(os.path.join(rootdir, subdir))

    return [path for path in paths if is_unhandled(path, handled_paths, excluded_paths)]

def is_unhandled(path, handled_paths, excluded_paths):
    return not any(path.startswith(excluded_path+'/') or path == excluded_path for excluded_path in excluded_paths) and path not in handled_paths

def handle_rename(rename, src, dest, env):
    config = unified_rename_config(rename)
    handled_paths = [config['path']]

    folder_path = '/'.join(config['path'].split('/')[:-1])

    src_path = f'{src}/{config["path"]}'
    dest_path = '/'.join((dest, folder_path, template_string(config['transform'], **without_excluded(env))))

    if src_path not in env['excluded']:

        if config['type'] == 'file':
            template_file(src_path, dest_path, **without_excluded(env))
            handled_paths.append(src_path)
        if config['type'] == 'folder':
            os.mkdir(dest_path)
            handled_paths.append(src_path)
            handled_paths += handle_node(config, src_path, dest_path, env)

    return handled_paths

def handle_loop(loop, src, dest, env):
    config = unified_loop_config(loop)
    handled_paths = [config['path']]

    context = benedict(env['context'])

    if config['var'] not in context:
        raise WrongConfigException(f'The variable {config["var"]} does not exist in the context.')

    src_path = f'{src}/{config["path"]}'
    local_env = lambda item: {
        'context': {config['item']: item, **env['context']},
        'filters': env['filters'],
        'tests': env['tests']
    }
    dest_path = lambda item: '/'.join((
        dest,
        '/'.join(config['path'].split('/')[:-1]),
        template_string(config['transform'], **local_env(item))
    ))

    if src_path not in env['excluded']:

        items = context[config['var']]
        for test in config['tests']:
            items = [item for item in items if env['tests'][test](item)]
        items = [item for item in items if item not in config['excluded_values']]
        for filter_ in config['filters']:
            items = env['filters'][filter_](items)

        if config['type'] == 'file':
            for item in items:
                template_file(src_path, dest_path(item), **local_env(item))
                handled_paths.append(src_path)
        if config['type'] == 'folder':
            for item in items:
                local_dest = dest_path(item)
                os.mkdir(local_dest)
                handled_paths.append(src_path)
                handled_paths += handle_node(config, src_path, local_dest, {**env, **local_env(item)})

    return handled_paths
